options nosource;

%* set SASCBTBL to where you placed FileList.api, examples;
* filename SASCBTBL 'F:\contrib\filelist\FileList.api';
* filename SASCBTBL CATALOG 'SASUSER.API.FILELIST.SOURCE';


%macro FileList (FileSpec, Out);
%******************************************************************************;
%* 4/20/98 Richard A. DeVenezia
%******************************************************************************;
%* Modifications:
%* 3/23/02 RAD Change +ms to +ms/1000.
%*             Thanks to Lex Jansen for finding this bug
%******************************************************************************;
%* Requires:
%*   Fileref SASCBTBL to point to location of FileList.api
%******************************************************************************;
%* Execution:
%*   This macro should only be invoked from submitted SAS code;
%******************************************************************************;
%* Arguments:
%*
%*   FileSpec - filename search string that can contain * and ?
%*     FileSpec can be considered as two parts, the path and the filename,
%*     <some-path>\<filename>. Only the filename part can contain the wildcards
%*     * and ?. A value like C:\*\*.SAS is INCORRECT, C:\MyStuff\*.SAS is CORRECT
%*     If you specify only a directory name then only the directory name
%*     (if it exists) will be found. You need to pad the directory name with
%*     \*.* to get a list of all files in a directory.
%*
%*   Out - Dataset to receive file information of matching filenames which will
%*         have the following variables:
%*     FTYPE    - directory or file
%*     FMOD     - modification date as SAS datetime
%*     FSIZE    - file size in bytes
%*     FNAME    - first 200 characters of filename (can be as long as 260)
%*     FNAME2   - last 60 characters of filename
%*                Filenames ending in spaces (a possibility but not very common)
%*                will NOT be recorded properly due to SAS character variable
%*                handling (placing spaces on end of character variables)
%*                FNAME and FNAME2 are SAS formatted to display least required
%*                characters. I.e. if longest fname is 20 characters then fname
%*                will be $CHAR20.
%*     FNAME83  - DOS 8.3 alternate filename (the one with tildes)
%*                (Users with NTFS partitions can turn off creation of 8.3
%*                 alternate names)
%******************************************************************************;
%* Note:
%*   Out will only contain the filenames, NOT the path to the filename;
%*   You must know the path to the filenames if you wish to use the Out
%*   dataset to reconstruct fully qualified filenames.
%******************************************************************************;
%* Examples:
%*   %FileList (C:\My Data Files\*.NEW, WORK.NEWLIST);
%*   %FileList (\\CorporateUNC\Lab Results\New England\XRAY*.DAT, WORK.XRYFILES);
%******************************************************************************;
%* Using the output dataset:
%*   * A common use of Out would be
%*   data results;
%*     set <Out>;
%*     where FTYPE='File';
%*     infile dummy filevar=fullname missover end=EOF;
%*     fullname = <PATH TO FILES> || '\' || FNAME;
%*     do while (not eof);
%*       input ...;
%*       ...
%*       output;
%*     end;
%*   run;
%******************************************************************************;
%* Tested environment:
%*   Windows NT 4.0 Service Pack 3
%******************************************************************************;
%* Why write this macro?:
%*   I wanted a flashless method of getting directory listings
%*
%*   X "Dir <path> > <outlist>", can make a file to read, but the screen
%*   flashes when the CMD window pops up, also, the exact modification time is
%*   not availale (dir does not list seconds)
%******************************************************************************;
%* What does the macro do?:
%*   Uses the Windows API functions FindFirstFile, FindNextFile and FindClose
%*   to obtain information about files matching a file specification.
%******************************************************************************;

  %****************************************************************************;
  %* Make sure SASCBTBL is assigned to something that exists;

  %local rc ;
  %let rc = %sysfunc (fileref(SASCBTBL));

  %if &rc > 0 %then %do;
    %put ERROR: Fileref SASCBTBL is not assigned;
    %goto EndMacro;
  %end;

  %if &rc < 0 %then %do;
    %put ERROR: Fileref SASCBTBL points to non-existent file;
    %goto EndMacro;
  %end;

  options source;

  %****************************************************************************;
  %* Search for the files indicated by the FileSpec;
  %* Note: Filenames that end with spaces will not be properly recorded;

  %local fnamef fname2f; %* will contain format lengths;
  %let fnamef=;
  %let fname2f=;

  data &OUT;
    retain ftype fmod fsize;
    length fname $200 fname2 $60 fname83 $14;

    * prevent uninitialized note;
    retain fattr created accessed modified sizeh sizel reserve0 reserve1 lmod 0;
    retain year month dow day h m s ms 0;

    handle = modulen ("FindFirstFileA", "&FILESPEC",
                      fattr, created, accessed, modified, sizeH, sizeL,
                      reserve0, reserve1,
                      fname, fname2, fname83);

    if handle ne -1 then do;
      do until (1 ne modulen ("FindNextFileA", handle,
                              fattr, created, accessed, modified, sizeH, sizeL,
                              reserve0, reserve1,
                              fname, fname2, fname83)
               );

        if index (fname, '00'x) then do;
          fname  = input (fname,$CSTR200.);
          fname2 = '';
        end;
        else
          fname2  = input (fname2,$CSTR200.);

        if (band (fattr, 10x))
           then ftype='Directory';
           else ftype='File';

        if fname83 = "" then fname83 = fname;

        fsize = 0ffffffffx * sizeH + sizeL;

        rc = modulen ("FileTimeToLocalFileTime", modified, lmod);
        rc = modulen ("FileTimeToSystemTime",
                      lmod,year,month,dow,day,h,m,s,ms);
        fmod = dhms ( mdy (month,day,year), h,m,s ) + ms/1000;

        maxlen  = max (maxlen, length(fname));
        maxlen2 = max (maxlen2,length(fname2));

        if fname not in ('.' '..')
        then
          OUTPUT;
      end;
      handle = modulen ("FindClose", handle);

      call symput ('fnamef',  "$"||trim(left(put(maxlen, 3.)))||".");
      call symput ('fname2f', "$"||trim(left(put(maxlen2,3.)))||".");
    end;
    else do;
      put 'No handle returned from FindFirstFile.';
    end;

    attrib
      fsize format=comma12.    informat=comma12.
      fmod  format=datetime21. informat=datetime21.
    ;

    keep ftype fmod fsize fname fname2 fname83;
  run;

  * assign the formats for variables fname and fname2;

  %local lib ds;
  %let out = &SYSLAST;
  %let ds = %scan(&OUT,2,%str(.));
  %if (&ds=) %then %do;
    %let lib = WORK;
    %let ds  = &out;
  %end;
  %else %do;
    %let lib = %scan(&OUT,1,%str(.));
  %end;

  proc datasets nolist lib=&lib;
    modify &DS;
    %if (&fnamef  ne) %then %str(format fname  &fnamef;);
    %if (&fname2f ne) %then %str(format fname2 &fname2f;);
  quit;

%EndMacro:

%mend FileList;

options source;
